@@ -41,6 +41,21 @@
 #include "pspg.h"
 #include "unicode.h"
 
+#include <ctype.h>
+
+#ifndef A_ITALIC
+#define A_ITALIC	A_DIM
+#endif
+
+static inline void
+wrepeatspace(WINDOW *win, int n)
+{
+	int		i;
+
+	for (i = 0; i < n; i++)
+		waddch(win, ' ');
+}
+
 /*
  * Flush data to window. When row is a decoration, then
  * replace ascii decoration by special terminal decoration.
@@ -215,7 +230,7 @@ print_column_names(WINDOW *win,
 
 		if (pos + chars > srcx)
 		{
-			wprintw(win, "%*s", pos + chars - srcx, "");
+			wrepeatspace(win, pos + chars - srcx);
 
 			pos += chars;
 			ptr += bytes;
@@ -276,11 +291,11 @@ print_column_names(WINDOW *win,
 			if (desc->linestyle == 'a' && opts->force_uniborder)
 				waddch(win, ACS_VLINE);
 			else
-				wprintw(win, "%.*s", bytes, ptr);
+				waddnstr(win, ptr, bytes);
 		}
 		else
 			/* clean background of colum names */
-			wprintw(win, "%*s", chars, "");
+			wrepeatspace(win, chars);
 
 		headline_ptr += chars;
 		ptr += bytes;
@@ -389,7 +404,7 @@ print_column_names(WINDOW *win,
 		}
 
 		wattron(win, new_attr);
-		mvwprintw(win, cy, act_xmin, "%.*s", colname_size, colname);
+		mvwaddnstr(win, cy, act_xmin, colname, colname_size);
 		wattroff(win, new_attr);
 	}
 }
@@ -417,7 +432,7 @@ set_line_info(Options *opts,
 			lb->lineinfo[i].mask = LINEINFO_UNKNOWN;
 	}
 
-	linfo = &lbm->lb->lineinfo[lbm->lb_rowno - 1];
+	linfo = &lbm->lb->lineinfo[lbm->lb_rowno];
 
 	if (linfo->mask & LINEINFO_UNKNOWN)
 	{
@@ -506,6 +521,28 @@ set_line_info(Options *opts,
 	return linfo;
 }
 
+#if NCURSES_WIDECHAR > 0 && defined HAVE_NCURSESW
+
+/*
+ * Two examples how to print wide char correctly
+ */
+static void
+pspg_mvwadd_wchar(WINDOW *win, int y, int x, wchar_t wchr, attr_t attr)
+{
+	cchar_t		cchr;
+
+	setcchar(&cchr, &wchr, attr, PAIR_NUMBER(attr), NULL);
+	mvwadd_wch(win, y, x, &cchr);
+}
+
+static void
+mvwadd_wchar(WINDOW *win, int y, int x, wchar_t wchr)
+{
+	mvwaddnwstr(win, y, x, &wchr, 1);
+}
+
+#endif
+
 /*
  * Draw scrollbar to related window.
  */
@@ -525,37 +562,56 @@ draw_scrollbar_win(WINDOW *win,
 		waddch(win, ACS_CKBOARD);
 
 	wattroff(win, t->scrollbar_attr);
-	wattron(win, t->scrollbar_arrow_attr);
+
+#if NCURSES_WIDECHAR > 0 && defined HAVE_NCURSESW
 
 	if (t->scrollbar_use_arrows)
 	{
 		if (!use_utf8 || opts->force_ascii_art)
 		{
+			wattron(win, t->scrollbar_arrow_attr);
+
 			mvwaddch(win, 0, 0, ACS_UARROW);
 			mvwaddch(win, scrdesc->scrollbar_maxy - 1, 0, ACS_DARROW);
+
+			wattroff(win, t->scrollbar_arrow_attr);
 		}
 		else
 		{
 			/* 🠕 🠗 */
-			mvwprintw(win, 0, 0, "%lc", L'\x1F815');
-			mvwprintw(win, scrdesc->scrollbar_maxy - 1, 0, "%lc", L'\x1F817');
+			wattron(win, t->scrollbar_arrow_attr);
+
+			mvwadd_wchar(win, 0, 0,  L'\x1F815');
+			mvwadd_wchar(win, scrdesc->scrollbar_maxy - 1, 0, L'\x1F817');
+
+			wattroff(win, t->scrollbar_arrow_attr);
 		}
 	}
 	else
 	{
 		/* ▲ ▼ */
-		mvwprintw(win, 0, 0, "%lc", L'\x25b2');
-		mvwprintw(win, scrdesc->scrollbar_maxy - 1, 0, "%lc", L'\x25bc');
+		pspg_mvwadd_wchar(win, 0, 0, L'\x25b2', t->scrollbar_arrow_attr);
+		pspg_mvwadd_wchar(win, scrdesc->scrollbar_maxy - 1, 0, L'\x25bc', t->scrollbar_arrow_attr);
 	}
 
+#else
+
+	wattron(win, t->scrollbar_arrow_attr);
+
+	mvwaddch(win, 0, 0, ACS_UARROW);
+	mvwaddch(win, scrdesc->scrollbar_maxy - 1, 0, ACS_DARROW);
+
 	wattroff(win, t->scrollbar_arrow_attr);
+
+#endif
+
 	wattron(win, scrdesc->scrollbar_mode ? t->scrollbar_active_slider_attr : t->scrollbar_slider_attr);
 
 	if (!t->scrollbar_slider_symbol)
 	{
 		/* draw slider */
 		for (i = 0; i < scrdesc->slider_size; i++)
-			mvwprintw(win, scrdesc->slider_min_y + i, 0, " ");
+			mvwaddch(win, scrdesc->slider_min_y + i, 0, ' ');
 	}
 	else
 		mvwaddch(win, scrdesc->slider_min_y, 0, t->scrollbar_slider_symbol);
@@ -595,6 +651,194 @@ is_in_searched_pattern(int pos,
 	return false;
 }
 
+typedef struct
+{
+	int		start_pos;
+	int		end_pos;
+	int		typ;
+} SpecialWord;
+
+static bool
+is_upper_char(char *chr)
+{
+	if (use_utf8)
+		return utf8_isupper(chr);
+
+	return isupper(*chr);
+}
+
+static bool
+is_ascii_alnum(char chr)
+{
+	if (use_utf8 && (chr & 0x80))
+		return false;
+
+	return isalnum(chr);
+}
+
+static bool
+is_ascii_alpha(char chr)
+{
+	if (use_utf8 && (chr & 0x80))
+		return false;
+
+	return isalpha(chr);
+}
+
+/*
+ * Try to identify words in line that should be highlighted
+ *
+ */
+static int
+parse_line(char *line, SpecialWord *words, int maxwords)
+{
+	int		nwords = 0;
+	int		pos = 0;
+	bool	first_nonspace = true;
+
+	/*
+	 * When text starts on line start and first char is upper
+	 * char, then mark this text until double colon.
+	 *
+	 * like "Usage:"
+	 */
+	if (is_upper_char(line) || strncmp(line, "psql", 4) == 0)
+	{
+		char   *aux_line = line;
+
+		words[0].start_pos = 0;
+		words[0].typ = 3;
+
+		while (*line != '\0' && *line != ':')
+		{
+			pos += dsplen(line);
+			line += charlen(line);
+		}
+
+		if (!is_upper_char(aux_line + 1))
+		{
+			words[0].end_pos = pos - 1;
+			words[0].typ = 3;
+			nwords = 1;
+			first_nonspace = false;
+		}
+		else
+		{
+			line = aux_line;
+			pos = 0;
+		}
+	}
+
+	while (*line)
+	{
+		while (*line == ' ')
+		{
+			line += 1;
+			pos += 1;
+		}
+
+		/* psql's backslash commands */
+		if (*line == '\\')
+		{
+			words[nwords].start_pos = pos;
+			words[nwords].typ = 1;
+
+			while (*line != ' ' && *line != '\0')
+			{
+				pos += dsplen(line);
+				line += charlen(line);
+			}
+		}
+		/* psql's shell options */
+		else if (*line == '-')
+		{
+			words[nwords].start_pos = pos;
+			words[nwords].typ = 1;
+
+			while (*line == '-')
+			{
+				line += 1;
+				pos += 1;
+			}
+
+			if (pos - words[nwords].start_pos > 2)
+				continue;
+
+			if (!is_ascii_alnum(*line))
+				continue;
+
+			while (is_ascii_alnum(*line) || *line == '-')
+			{
+				line += 1;
+				pos += 1;
+			}
+
+			words[nwords].typ = 1;
+			goto fin;
+		}
+		/* psql's variables and \pset variables */
+		else if (is_ascii_alpha(*line))
+		{
+			bool	only_upper = true;
+			char   *start = line;
+
+			words[nwords].start_pos = pos;
+
+			while (is_ascii_alnum(*line) || *line == '_')
+			{
+				if (islower(*line))
+					only_upper = false;
+
+				line += 1;
+				pos += 1;
+			}
+
+			if (!only_upper)
+			{
+				if (first_nonspace && words[nwords].start_pos == 2 && *line == '\0')
+				{
+					words[nwords].typ = 2;
+					goto fin;
+				}
+				else if (isupper(*start) && *line == ':')
+				{
+					words[nwords].typ = 3;
+					goto fin;
+				}
+
+				continue;
+			}
+
+			if (strncmp(start, "SQL", pos - words[nwords].start_pos) == 0)
+				continue;
+
+			if (*line == ':')
+				continue;
+
+			if ((pos - words[nwords].start_pos) == 1)
+				continue;
+
+			words[nwords].typ = 2;
+		}
+		else
+		{
+			pos += dsplen(line);
+			line += charlen(line);
+			first_nonspace = false;
+			continue;
+		}
+
+fin:
+
+		words[nwords].end_pos = pos - 1;
+
+		if (++nwords == maxwords)
+			return nwords;
+	}
+
+	return nwords;
+}
+
 void
 window_fill(int window_identifier,
 			int srcy,
@@ -618,6 +862,8 @@ window_fill(int window_identifier,
 	char		*free_row;
 	WINDOW		*win;
 	Theme		*t;
+	SpecialWord specwords[30];
+	int			nspecwords;
 
 	bool		is_footer = window_identifier == WINDOW_FOOTER;
 	bool		is_fix_rows = window_identifier == WINDOW_LUC || window_identifier == WINDOW_FIX_ROWS;
@@ -630,12 +876,29 @@ window_fill(int window_identifier,
 								window_identifier == WINDOW_FIX_COLS ||
 								window_identifier == WINDOW_FIX_ROWS ||
 								window_identifier == WINDOW_FOOTER;
+	bool		is_text = window_identifier == WINDOW_FOOTER && desc->headline_transl == NULL;
+	int			odd_theme_identifier = -1;
 
 	win = scrdesc->wins[window_identifier];
 	t = &scrdesc->themes[window_identifier];
 
 	pattern_fix = t->found_str_attr & A_UNDERLINE;
 
+	if (has_odd_themedef && opts->highlight_odd_rec)
+	{
+		if (window_identifier == WINDOW_FIX_COLS)
+			odd_theme_identifier = WINDOW_FIX_COLS_ODD;
+		else if (window_identifier == WINDOW_ROWS)
+			odd_theme_identifier = WINDOW_ROWS_ODD;
+		else if (window_identifier == WINDOW_ROWNUM)
+			odd_theme_identifier = WINDOW_ROWNUM_ODD;
+		else
+			odd_theme_identifier = -1;
+
+		if (odd_theme_identifier != -1)
+			multilines_detection(desc);
+	}
+
 	/* when we want to detect expanded records titles */
 	if (desc->is_expanded_mode)
 	{
@@ -682,12 +945,24 @@ window_fill(int window_identifier,
 		int			npositions = 0;
 		int			is_in_range = false;
 		int			rowno = row + srcy_bak + 1 - desc->first_data_row;
+		int			lineno;
+		int			recno;
 
 		is_cursor_row = (!opts->no_cursor && row == cursor_row);
 
 		(void) lbi_set_mark_next(&lbi, &lbm);
 
-		line_is_valid = lbm_get_line(&lbm, &rowstr, &lineinfo, NULL);
+		line_is_valid = lbm_get_line(&lbm, &rowstr, &lineinfo, &lineno);
+
+		if (odd_theme_identifier != -1)
+		{
+			recno = lineno - lineinfo->recno_offset;
+
+			if (recno % 2 == 1)
+				t = &scrdesc->themes[odd_theme_identifier];
+			else
+				t = &scrdesc->themes[window_identifier];
+		}
 
 		/* when rownum is printed, don't process original text */
 		if (is_rownum && line_is_valid)
@@ -813,6 +1088,11 @@ window_fill(int window_identifier,
 			bool	is_bottom_deco = false;
 			int		trailing_spaces = 0;
 
+			if (is_text)
+				nspecwords = parse_line(rowstr, specwords, 30);
+			else
+				nspecwords = 0;
+
 			is_top_deco = effective_row == desc->border_top_row;
 			is_head_deco = effective_row == desc->border_head_row;
 			is_bottom_deco = effective_row == desc->border_bottom_row;
@@ -914,6 +1194,17 @@ window_fill(int window_identifier,
 			/* find length of maxx characters */
 			if (*ptr != '\0')
 			{
+				bool	is_selected_rows;
+				bool	is_selected_row;
+				bool	is_selected_columns;
+				int		specword_typ;
+
+				is_selected_rows = is_selectable && scrdesc->selected_first_row != -1;
+				is_selected_row = rowno >= scrdesc->selected_first_row + 1 &&
+								  rowno < scrdesc->selected_first_row + 1 + scrdesc->selected_rows;
+
+				is_selected_columns = is_selectable && selected_xmin != -1;
+
 				while (i < maxx)
 				{
 					bool	is_cursor;
@@ -921,13 +1212,27 @@ window_fill(int window_identifier,
 					bool	is_vertical_cursor = false;
 					int		pos = (i != -1) ? srcx + i : -1;
 					bool	skip_char = false;
+					specword_typ = 0;
+
+					if (nspecwords > 0)
+					{
+						int		j;
+
+						for (j = 0; j < nspecwords; j++)
+						{
+							if (pos >= specwords[j].start_pos && pos <= specwords[j].end_pos)
+							{
+								specword_typ = specwords[j].typ;
+								break;
+							}
+						}
+					}
 
 					is_in_range = false;
 
-					if (is_selectable && scrdesc->selected_first_row != -1)
+					if (is_selected_rows)
 					{
-						if (rowno >= scrdesc->selected_first_row + 1 &&
-							rowno < scrdesc->selected_first_row + 1 + scrdesc->selected_rows)
+						if (is_selected_row)
 						{
 							if (selected_xmin != -1 && pos != -1)
 							{
@@ -938,7 +1243,7 @@ window_fill(int window_identifier,
 								is_in_range = true;
 						}
 					}
-					else if (is_selectable && selected_xmin != -1 && pos != -1)
+					else if (is_selected_columns && pos != -1)
 					{
 						if (pos >= selected_xmin && pos <= selected_xmax)
 							is_in_range = true;
@@ -1044,6 +1349,11 @@ window_fill(int window_identifier,
 								new_attr = column_format == 'd' ? t->data_attr : t->line_attr;
 						}
 
+						if (specword_typ == 1 || specword_typ == 2)
+							new_attr |= A_BOLD;
+						else if (specword_typ == 3)
+							new_attr |= A_ITALIC | A_UNDERLINE;
+
 						if (is_cursor || is_cross_cursor)
 						{
 							if (is_found_row && pos >= scrdesc->found_start_x &&
@@ -1188,7 +1498,7 @@ window_fill(int window_identifier,
 			/* print trailing spaces necessary for correct vertical cursor */
 			if (trailing_spaces > 0)
 			{
-				wprintw(win, "%*s", trailing_spaces, "");
+				wrepeatspace(win, trailing_spaces);
 				i += trailing_spaces;
 			}
 
@@ -1481,7 +1791,7 @@ draw_data(Options *opts, ScrDesc *scrdesc, DataDesc *desc,
 	struct winsize size;
 	int		i;
 
-	if (ioctl(0, TIOCGWINSZ, (char *) &size) >= 0)
+	if (ioctl(fileno(stdout), TIOCGWINSZ, (char *) &size) >= 0)
 	{
 		int		expected_rows = min_int(size.ws_row - 1 - scrdesc->top_bar_rows,
 										desc->last_row + 1);
