@@ -14,12 +14,14 @@
 #include <ctype.h>
 #include <errno.h>
 #include <libgen.h>
+#include <poll.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
 
+#include "inputs.h"
 #include "pspg.h"
 #include "unicode.h"
 
@@ -28,19 +30,6 @@
 #include <time.h>
 #include <math.h>
 
-
-static void
-current_time(time_t *sec, long *ms)
-{
-	struct timespec spec;
-
-	clock_gettime(CLOCK_MONOTONIC, &spec);
-	*ms = roundl(spec.tv_nsec / 1.0e6);
-	*sec = spec.tv_sec;
-}
-
-#define time_diff(s1, ms1, s2, ms2)		((s1 - s2) * 1000 + ms1 - ms2)
-
 static void
 print_duration(time_t start_sec, long start_ms, const char *label)
 {
@@ -83,10 +72,14 @@ isHeadLeftChar(char *str)
 {
 	const char *u1 = "\342\224\200";
 	const char *u2 = "\342\225\220";
-	const char *u3 = "\342\225\236";
-	const char *u4 = "\342\224\234";
-	const char *u5 = "\342\225\240";
-	const char *u6 = "\342\225\237";
+
+	const char *u3 = "\342\225\236"; /* ╞ */
+	const char *u4 = "\342\224\234"; /* ├ */
+	const char *u5 = "\342\225\240"; /* ╠ */
+	const char *u6 = "\342\225\237"; /* ╟ */
+
+	const char *u7 = "\342\224\214"; /* ┌ */
+	const char *u8 = "\342\225\224"; /* ╔ */
 
 	/* ascii */
 	if ((str[0] == '+' || str[0] == '-') && str[1] == '-')
@@ -108,6 +101,13 @@ isHeadLeftChar(char *str)
 	if (str[0] == '+' && str[1] == '=')
 		return true;
 
+	/* Oracle's SQLcli (ANSI CONSOLE) */
+	if (str[0] == '_')
+		return true;
+
+	if (str[0] != '\342')
+		return false;
+
 	if (strncmp(str, u1, 3) == 0)
 		return true;
 	if (strncmp(str, u2, 3) == 0)
@@ -120,6 +120,39 @@ isHeadLeftChar(char *str)
 		return true;
 	if (strncmp(str, u6, 3) == 0)
 		return true;
+	if (strncmp(str, u7, 3) == 0)
+		return true;
+	if (strncmp(str, u8, 3) == 0)
+		return true;
+
+	return false;
+}
+
+static bool
+isUnicodeHeadLeftCharBorder2(char *str)
+{
+	const char *u3 = "\342\225\236"; /* ╞ */
+	const char *u4 = "\342\224\234"; /* ├ */
+	const char *u5 = "\342\225\240"; /* ╠ */
+	const char *u6 = "\342\225\237"; /* ╟ */
+	const char *u7 = "\342\224\214"; /* ┌ */
+	const char *u8 = "\342\225\224"; /* ╔ */
+
+	if (str[0] != '\342')
+		return false;
+
+	if (strncmp(str, u3, 3) == 0)
+		return true;
+	if (strncmp(str, u4, 3) == 0)
+		return true;
+	if (strncmp(str, u5, 3) == 0)
+		return true;
+	if (strncmp(str, u6, 3) == 0)
+		return true;
+	if (strncmp(str, u7, 3) == 0)
+		return true;
+	if (strncmp(str, u8, 3) == 0)
+		return true;
 
 	return false;
 }
@@ -182,21 +215,24 @@ is_expanded_header(char *str, int *ei_minx, int *ei_maxx)
 	if (strncmp(str, "[ ", 2) != 0)
 		return false;
 
-	if (ei_minx != NULL && ei_maxx != NULL)
-	{
-		pos += 2;
-		str += 2;
+	pos += 2;
+	str += 2;
+
+	if (ei_minx)
 		*ei_minx = pos - 1;
 
-		while (*str != ']' && *str != '\0')
-		{
-			pos += 1;
-			str += charlen(str);
-		}
-
-		*ei_maxx = pos - 1;
+	while (*str != ']' && *str != '\0')
+	{
+		pos += 1;
+		str += charlen(str);
 	}
 
+	if (strncmp(str - 1, " ]", 2) != 0)
+		return false;
+
+	if (ei_maxx)
+		*ei_maxx = pos - 1;
+
 	return true;
 }
 
@@ -354,12 +390,12 @@ is_cmdtag(char *str)
 #define STATBUF_SIZE		(10 * 1024)
 
 static size_t
-_getline(char **lineptr, size_t *n, FILE *fp, bool is_blocking, bool wait_on_data)
+_getline(char **lineptr, size_t *n, FILE *fp, bool is_nonblocking, bool wait_on_data)
 {
 	int			_errno;
 	ssize_t		result;
 
-	if (is_blocking)
+	if (!is_nonblocking)
 	{
 		result = getline(lineptr, n, fp);
 		_errno = errno;
@@ -384,7 +420,7 @@ _getline(char **lineptr, size_t *n, FILE *fp, bool is_blocking, bool wait_on_dat
 
 		for (;;)
 		{
-			char   *str;
+			char	*str;
 			int	len = 0;
 
 			errno = 0;
@@ -400,9 +436,17 @@ _getline(char **lineptr, size_t *n, FILE *fp, bool is_blocking, bool wait_on_dat
 
 				if (dynbuf)
 				{
-					dynbuf = realloc(dynbuf, fetched_chars + len + 1);
-					if (!dynbuf)
+					char	   *_dynbuf;
+
+					_dynbuf = realloc(dynbuf, fetched_chars + len + 1);
+					if (!_dynbuf)
+					{
+						free(_dynbuf);
 						return -1;
+					}
+					else
+						dynbuf = _dynbuf;
+
 					memcpy(dynbuf + fetched_chars, statbuf, len + 1);
 					fetched_chars += len;
 				}
@@ -437,13 +481,9 @@ endline_exit:
 				errno = _errno;
 			}
 
-			if (errno)
+			if (errno || feof(fp))
 			{
-				if (feof(fp))
-				{
-					goto endline_exit;
-				}
-				else if (errno == EAGAIN)
+				if (errno == EAGAIN)
 				{
 					struct pollfd fds[1];
 
@@ -457,13 +497,29 @@ endline_exit:
 					if (rc == -1)
 					{
 						log_row("poll error (%s)",  strerror(errno));
+						if (handle_sigint)
+						{
+							free(dynbuf);
+							handle_sigint = false;
+							return -1;
+						}
 
 						usleep(1000);
 					}
 
+					if (fds[0].revents & POLLHUP)
+					{
+						free(dynbuf);
+						return -1;
+					}
+
 					clearerr(fp);
 					continue;
 				}
+				else if (feof(fp))
+				{
+					goto endline_exit;
+				}
 				else
 				{
 					free(dynbuf);
@@ -529,6 +585,81 @@ strncpytrim(char *dest, const char *src,
 	*dest = '\0';
 }
 
+/*
+ * Remove ANSI escape sequences from input line
+ */
+static ssize_t
+remove_ansi_escape_seq(char *line, ssize_t bytes)
+{
+	char	   *ptr = line;
+	int			not_processed = bytes;
+	char	   *writeptr = ptr;
+
+	if (bytes == 0)
+		return 0;
+
+	/* fast mode, escape detection */
+	while (not_processed > 0)
+	{
+		if (*ptr == '\x1b')
+			break;
+
+		not_processed -= 1;
+		ptr += 1;
+	}
+
+	/* when there are not escape sequence, then returns original size */
+	if (not_processed == 0)
+		return bytes;
+
+	writeptr = ptr;
+
+	/*
+	 * there are ANSI escape sequences, and
+	 * we need to clean it.
+	 */
+	while (not_processed > 0)
+	{
+		if (*ptr != '\x1b')
+		{
+			*writeptr++ = *ptr++;
+			not_processed -= 1;
+		}
+		else
+		{
+			if (not_processed > 2)
+			{
+				not_processed -= 2;
+				ptr += 1;
+
+				if (*ptr++ == '[')
+				{
+					while (not_processed-- > 0)
+					{
+						int		c = *ptr++;
+
+						if (c)
+						{
+							if (c >= '@' && c <= '~')
+								break;
+						}
+						else
+							/* broken ascii escape sequence or end of line? */
+							goto finish_deescape;
+					}
+				}
+			}
+			else
+				break;
+		}
+	}
+
+finish_deescape:
+
+	*writeptr++ = '\0';
+
+	return writeptr - line;
+}
 
 /*
  * Read data from file and fill DataDesc.
@@ -540,7 +671,12 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 	size_t		len;
 	ssize_t		read;
 	int			nrows = 0;
+	int			stop_after_nrows = 0;
+	bool		completed = true;
+	bool		initial_run;
+	bool		progressive_load_mode;
 	LineBuffer *rows;
+	int		clen = -1;
 
 #ifdef DEBUG_PIPE
 
@@ -552,35 +688,58 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 
 #endif
 
-	desc->title[0] = '\0';
-	desc->title_rows = 0;
-	desc->border_top_row = -1;
-	desc->border_head_row = -1;
-	desc->border_bottom_row = -1;
-	desc->first_data_row = -1;
-	desc->last_data_row = -1;
-	desc->is_expanded_mode = false;
-	desc->headline_transl = NULL;
-	desc->cranges = NULL;
-	desc->columns = 0;
-	desc->footer_row = -1;
-	desc->alt_footer_row = -1;
-	desc->is_pgcli_fmt = false;
-	desc->namesline = NULL;
-	desc->order_map = NULL;
-	desc->total_rows = 0;
+	progressive_load_mode = opts->progressive_load_mode;
 
-	desc->maxbytes = -1;
-	desc->maxx = -1;
+	if (!desc->initialized)
+	{
+		log_row("DataDesc is initialized\n");
 
-	memset(&desc->rows, 0, sizeof(LineBuffer));
-	rows = &desc->rows;
-	desc->rows.prev = NULL;
-	desc->oid_name_table = false;
-	desc->multilines_already_tested = false;
+		desc->title[0] = '\0';
+		desc->title_rows = 0;
+		desc->border_top_row = -1;
+		desc->border_head_row = -1;
+		desc->border_bottom_row = -1;
+		desc->first_data_row = -1;
+		desc->last_data_row = -1;
+		desc->is_expanded_mode = false;
+		desc->headline_transl = NULL;
+		desc->cranges = NULL;
+		desc->columns = 0;
+		desc->footer_row = -1;
+		desc->alt_footer_row = -1;
+		desc->is_pgcli_fmt = false;
+		desc->namesline = NULL;
+		desc->order_map = NULL;
+		desc->total_rows = 0;
+		desc->load_data_rows = false;
+
+		desc->maxbytes = -1;
+		desc->maxx = -1;
+
+		memset(&desc->rows, 0, sizeof(LineBuffer));
+		desc->rows.prev = NULL;
+		desc->freeze_two_cols = false;
+		desc->multilines_already_tested = false;
+		desc->last_buffer = 0;
+
+		/* safe reset */
+		desc->filename[0] = '\0';
+
+		/* progress load mode */
+		desc->initialized = true;
+		desc->completed = false;
+	}
+
+	nrows = desc->total_rows;
+
+	/*
+	 * DataDesc struct can be copied, and then has not
+	 * contain any self reference. So instead to set
+	 * last_buffer to &desc->rows, I use NULL as signal,
+	 * so no other line buffer was created (and used).
+	 */
+	rows = desc->last_buffer ? desc->last_buffer : &desc->rows;
 
-	/* safe reset */
-	desc->filename[0] = '\0';
 	state->errstr = NULL;
 	state->_errno = 0;
 
@@ -593,44 +752,63 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 		desc->filename[64] = '\0';
 	}
 
-	if (!state->fp)
+	if (!f_data)
 		return false;
 
-	clearerr(state->fp);
+	clearerr(f_data);
 
-	/* detection truncating */
-	if (state->detect_truncation)
+	if (progressive_load_mode)
 	{
-		struct stat stats;
-
-		if (fstat(fileno(state->fp), &stats) == 0)
-		{
-			if (stats.st_size < state->last_position)
-			{
-				log_row("file \"%s\" was truncated", opts->pathname);
-				fseek(state->fp,0L, SEEK_SET);
-			}
-		}
+		if (nrows == 0)
+			stop_after_nrows = max_int(2 * LINES, 500);
 		else
-			log_row("cannot to stat file: %s (%s)", opts->pathname, strerror(errno));
+			stop_after_nrows = nrows + 2000;
+	}
+	else
+	{
+		stop_after_nrows = -1;
+		desc->completed = true;
 	}
 
+	if (nrows == 0)
+	{
+		/* detection truncating */
+		detect_file_truncation();
+		initial_run = true;
+	}
+	else
+		initial_run = false;
+
 	errno = 0;
-	read = _getline(&line, &len, state->fp, state->is_blocking, false);
+	read = _getline(&line, &len, f_data, f_data_opts & STREAM_IS_IN_NONBLOCKING_MODE, false);
 	if (read == -1)
 		return false;
 
 	do
 	{
-		int		clen;
-
-		if (line && line[read - 1] == '\n')
+		if (line && read > 0 && line[read - 1] == '\n')
 		{
 			line[read - 1] = '\0';
 			read -= 1;
 		}
 
-		/* In streaming mode exit when you find empty row */
+		/*
+		 * In streaming mode go out when you find empty row.
+		 *
+		 * Attention: streaming mode can be used only for tabular data!!!
+		 * on nontabular data (plain text) we have not a possibility
+		 * to detect end of block. In theory we can wait sone time to
+		 * data, and after timeout we can alert end of block, but it
+		 * increase the time of data load (and complexity).
+		 *
+		 * So until we get more possibilities (some marks in stream),
+		 * it is better to hold stream mode as special case. I had an
+		 * idea to set stream mode as default, but it doesn't work with
+		 * non tabular data.
+		 *
+		 * Note: psql helps with it - it redirects only tabular data.
+		 *
+		 */
 		if (state->stream_mode && read == 0)
 		{
 			free(line);
@@ -642,6 +820,8 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 			break;
 		}
 
+		read = remove_ansi_escape_seq(line, read);
+
 		/* In query stream node exit when you find row with only GS - Group Separator */
 		if (opts->querystream && read == 1)
 		{
@@ -649,7 +829,13 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 				break;
 		}
 
-		clen = use_utf8 ? utf_string_dsplen(line, read) : read;
+		/*
+		 * When Unicode border 2 is used, then we can save CPU cycles,
+		 * becase we can very well detect begin and end of table. Inside
+		 * the table we don't need to check display width.
+		 */
+		if (clen == -1 || !desc->load_data_rows)
+			clen = use_utf8 ? utf_string_dsplen(line, read) : read;
 
 		if (rows->nrows == LINEBUFFER_LINES)
 		{
@@ -683,11 +869,19 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 		{
 			desc->border_top_row = nrows;
 			desc->is_expanded_mode = is_expanded_header(line, NULL, NULL);
+			if (desc->is_expanded_mode)
+				desc->border_head_row = nrows;
 		}
 		else if (desc->border_head_row == -1 && isHeadLeftChar(line))
 		{
 			desc->border_head_row = nrows;
 
+			if (isUnicodeHeadLeftCharBorder2(line))
+			{
+				desc->load_data_rows = true;
+				log_row("next row will be data row");
+			}
+
 			if (!desc->is_expanded_mode)
 				desc->is_expanded_mode = is_expanded_header(line, NULL, NULL);
 
@@ -703,6 +897,9 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 		{
 			desc->border_bottom_row = nrows;
 			desc->last_data_row = nrows - 1;
+			desc->load_data_rows = false;
+			log_row("next row will be desc row");
+
 		}
 		else if (!desc->is_expanded_mode && desc->border_bottom_row != -1 && desc->footer_row == -1)
 		{
@@ -713,6 +910,7 @@ readfile(Options *opts, DataDesc *desc, StateData *state)
 			/* Outer border is repeated in expanded mode, use last detected row */
 			desc->border_bottom_row = nrows;
 			desc->last_data_row = nrows - 1;
+			log_row("next row will be desc row");
 		}
 
 		if (!desc->is_expanded_mode && desc->border_head_row != -1 && desc->border_head_row < nrows
@@ -741,9 +939,27 @@ next_row:
 
 		line = NULL;
 
-		read = _getline(&line, &len, state->fp, state->is_blocking, true);
+		if (stop_after_nrows > 0 && nrows >= stop_after_nrows)
+		{
+			completed = false;
+			log_row("progressive load stop on %d row", nrows);
+			break;
+		}
+
+		if ((f_data_opts & STREAM_HAS_NOTIFY_SUPPORT) &&
+				nrows % 1000 == 0)
+		{
+			log_row("sleep 10ms per 1000 rows");
+			usleep(1000 * 10);
+		}
+
+		read = _getline(&line, &len, f_data, f_data_opts & STREAM_IS_IN_NONBLOCKING_MODE, true);
 	} while (read != -1);
 
+	desc->total_rows = nrows;
+	desc->last_buffer = rows != &desc->rows ? rows : NULL;
+	desc->completed = completed;
+
 	if (errno && errno != EAGAIN)
 	{
 		log_row("cannot to read from file (%s)", strerror(errno));
@@ -751,28 +967,29 @@ next_row:
 		return false;
 	}
 
-	if (state->detect_truncation)
-		state->last_position = ftell(state->fp);
-
-	desc->total_rows = nrows;
+	/* used for file truncation detection */
+	save_file_position();
 
 	log_row("read rows %d", nrows);
 
-	/*
-	 * border headline cannot be higher than 1000, to simply find it
-	 * in first row block. Higher number is surelly wrong, probably
-	 * some comment.
-	 */
-	if (desc->border_top_row >= 1000)
-		desc->border_top_row = -1;
-	if (desc->border_head_row >= 1000)
-		desc->border_head_row = -1;
+	if (initial_run)
+	{
+		/*
+		 * border headline cannot be higher than 1000, to simply find it
+		 * in first row block. Higher number is surelly wrong, probably
+		 * some comment.
+		 */
+		if (desc->border_top_row >= 100)
+			desc->border_top_row = -1;
+		if (desc->border_head_row >= 100)
+			desc->border_head_row = -1;
+
+		desc->headline_char_size = 0;
+	}
 
 	if (desc->last_row != -1)
 		desc->maxy = desc->last_row;
 
-	desc->headline_char_size = 0;
-
 	if (desc->border_head_row != -1)
 	{
 		if (desc->border_head_row == 0 && !desc->is_expanded_mode)
@@ -790,13 +1007,18 @@ next_row:
 
 		if (desc->border_head_row >= 1)
 			desc->namesline = desc->rows.rows[desc->border_head_row - 1];
-
 	}
 	else if (desc->is_expanded_mode && desc->border_top_row != -1)
 	{
 		desc->headline = desc->rows.rows[desc->border_top_row];
 		desc->headline_size = strlen(desc->headline);
 	}
+	else if (desc->border_head_row == -1 && desc->border_top_row != -1)
+	{
+		desc->border_head_row = desc->border_top_row;
+		desc->headline = desc->rows.rows[desc->border_top_row];
+		desc->headline_size = strlen(desc->headline);
+	}
 	else
 	{
 broken_format:
@@ -817,9 +1039,14 @@ broken_format:
 
 #endif
 
-	/* clean event buffer */
-	if (state->inotify_fd >= 0)
-		lseek(state->inotify_fd, 0, SEEK_END);
+#if defined(HAVE_INOTIFY) || defined(HAVE_KQUEUE)
+
+
+	if (completed && f_data_opts & STREAM_HAS_NOTIFY_SUPPORT) /* clean event buffer */
+		clean_notify_poll();
+
+#endif
+
 
 	return true;
 }
@@ -836,6 +1063,8 @@ translate_headline(DataDesc *desc)
 	bool	broken_format = false;
 	int		processed_chars = 0;
 	bool	is_expanded_info = false;
+	bool	is_headerless = (desc->border_head_row == desc->border_top_row &&
+							 !desc->is_expanded_mode);
 
 	srcptr = desc->headline;
 	destptr = smalloc(desc->headline_size + 2);
@@ -890,7 +1119,8 @@ translate_headline(DataDesc *desc)
 				  strncmp(srcptr, "\342\225\224", 3) == 0)   /* ╔ */
 		{
 			/* should be expanded mode */
-			if (processed_chars > 0 || !desc->is_expanded_mode)
+			if (processed_chars > 0 ||
+				(!desc->is_expanded_mode && !is_headerless))
 			{
 				broken_format = true;
 				break;
@@ -904,7 +1134,7 @@ translate_headline(DataDesc *desc)
 				 strncmp(srcptr, "\342\225\227", 3) == 0)   /* ╗ */
 		{
 			if (desc->linestyle != 'u' || desc->border_type != 2 ||
-				!desc->is_expanded_mode)
+				(!desc->is_expanded_mode  && !is_headerless))
 			{
 				broken_format = true;
 				break;
@@ -917,7 +1147,8 @@ translate_headline(DataDesc *desc)
 				 strncmp(srcptr, "\342\225\245", 3) == 0 || /* ╥╦ */
 				 strncmp(srcptr, "\342\225\246", 3) == 0)
 		{
-			if (desc->linestyle != 'u' || !desc->is_expanded_mode)
+			if (desc->linestyle != 'u' ||
+				(!desc->is_expanded_mode  && !is_headerless))
 			{
 				broken_format = true;
 				break;
@@ -1022,6 +1253,21 @@ translate_headline(DataDesc *desc)
 			*destptr++ = 'd';
 			srcptr += 1;
 		}
+		else if (*srcptr == '_')
+		{
+			if (processed_chars == 0)
+			{
+				desc->linestyle = 'a';
+				desc->border_type = 0;
+			}
+			else if (desc->linestyle != 'a')
+			{
+				broken_format = true;
+				break;
+			}
+			*destptr++ = 'd';
+			srcptr += 1;
+		}
 		else if (*srcptr == '|')
 		{
 			if (processed_chars == 0 && srcptr[1] == '-')
@@ -1187,7 +1433,20 @@ translate_headline(DataDesc *desc)
 				if (desc->cranges[1].name_size > 4 &&
 						nstrstr_with_sizes(desc->namesline + desc->cranges[1].name_offset + desc->cranges[1].name_size - 4,
 										   4, "name", 4))
-					desc->oid_name_table = true;
+					desc->freeze_two_cols = true;
+			}
+			else if ((desc->cranges[0].name_size == 10 &&
+						  				nstrstr_with_sizes(desc->namesline + desc->cranges[0].name_offset,
+									   desc->cranges[0].name_size,
+									   "schemaname",
+									   10)) ||
+					(desc->cranges[0].name_size == 6 &&
+						  				nstrstr_with_sizes(desc->namesline + desc->cranges[0].name_offset,
+									   desc->cranges[0].name_size,
+									   "schema",
+									   6)))
+			{
+				desc->freeze_two_cols = true;
 			}
 		}
 
@@ -1522,6 +1781,7 @@ multilines_detection(DataDesc *desc)
 {
 	LineBufferIter	lbi;
 	LineBufferMark	lbm;
+	int				recno = 1;
 
 	bool		border0 = (desc->border_type == 0);
 	bool		border1 = (desc->border_type == 1);
@@ -1587,6 +1847,11 @@ multilines_detection(DataDesc *desc)
 			pos += dsplen(str);
 			str += charlen(str);
 		}
+
+		lbm_recno_offset(&lbm, lineno - recno);
+
+		if (!found_continuation_symbol)
+			recno += 1;
 	}
 
 	desc->multilines_already_tested = true;
